# jahplayer - A widget library representation.

# Copyright (C) 2006 VM Inc.
# Released under the GPL.
# For more information, see http://www.jahshaka.org.

import openassetlib
import openpluginlib
import openimagelib
import jahwidgets
import jahstate
import utilities
import assets
import string
import sys
import os.path

#
# Shared handler object
#

class shared_handler:
	def __init__( self, reel ):
		self.reel = reel
		self.state = reel.state
		self.player = reel.player
		self.scribbler = jahwidgets.scribbler( utilities.get_absolute_path( "jahplayer-font.png" ), 16, 16, 3 )
		self.mouse_timer = jahwidgets.timer( )
		self.render_timer = jahwidgets.timer( )

#
# Timer delegate for moving stuff
#

class timed_mover( jahwidgets.timer_delegate ):
	def __init__( self, handler ):
		jahwidgets.timer_delegate.__init__( self )
		self.handler = handler

	def timer( self ):
		if self.handler.diff_x:
			self.handler.offset = self.handler.offset + 4 * self.handler.diff_x
			self.handler.shared.player.refresh( )
		return self.handler.diff_x != 0

#
# Reel mode handler
#

class reel_handler:
	# Constructor
	def __init__( self, shared ):
		self.shared = shared
		self.jog_x = 0
		self.diff_x = 0
		self.moving = None
		self.timed_mover = timed_mover( self )
		self.offset = 0

	def activate( self ):
		self.shared.state.player.loop = True
		self.shared.player.clear( )
		reel = self.shared.state.get_active_reel( )
		for i in reel.set:
			media = reel.get_media( guid = i.guid( ) )
			if media is not None:
				self.shared.state.prepare_media( media )
				self.shared.player.push_back( media )
		self.shared.player.refresh( )

	def accept_drop( self, event ):
		return True

	def process_drop( self, event ):
		self.shared.state.process_media( event )

	def notify_did_resize( self, w, h ):
		pass

	def media_needed( self, media ):
		self.calculate_bounds( )
		reel = media.reel
		index = reel.find( media )
		x = ( index + 1 ) * ( self.shared.reel.current_w + self.shared.reel.gutter ) 
		return x + self.offset >= 0 and x + self.offset - self.shared.reel.current_w < self.shared.player.width

	# Place the media item in the reel
	def media_geometry( self, media, width, height ):
		# Calculate the bound width and height
		image = media.image
		width, height = int( image.height( ) * media.ar( ) ), image.height( )
		width, height = utilities.bind_size_to( width, height, self.shared.reel.max_w, self.shared.reel.max_h )

		# Obtain the current reel
		reel = media.reel

		if media.dirty and media.image is not None and reel.get_thumb( media ) is None:
			media.image = self.shared.scribbler.rescale( media.image, width, height )
			reel.set_thumb( media.image, media )
		else:
			thumb = reel.get_thumb( media )
			width = thumb.width( )
			height = thumb.height( )

		# Locate the index of the media in the list
		index = reel.find( media )

		# Show everything except the active media as semi-transparent
		media.t = 0.5
		if self.shared.state.get_active_media( ) is media: media.t = 1.0

		# Centre the image on the horizontal and vertical
		width, height = utilities.bind_size_to( width, height, self.shared.reel.current_w, self.shared.reel.current_h )
		x = index * ( self.shared.reel.current_w + self.shared.reel.gutter ) + ( self.shared.reel.current_w - width ) / 2
		y = ( self.shared.reel.current_h - height ) / 2

		# Position the media relative to the offset
		media.set_geometry( self.offset + x, y, width, height )

	# Allows caching
	def media_rendered( self, media ):
		# Free up resources for all but the active media item
		if not media.dirty:
			media.close( )

	def mouse_press_event( self, event ):
		event.accept( )
		if event.isDoubleClick:
			media = self.shared.player.find( event.x, event.y )
			self.shared.state.set_active_media( media )
		self.jog_x = event.x
		self.diff_x = 0

	def mouse_move_event( self, event ):
		event.accept( )
		if self.jog_x:
			if event.x - self.jog_x:
				self.diff_x = event.x - self.jog_x
			self.offset = self.offset + self.diff_x
			if event.x > 0: self.jog_x = event.x
			self.shared.player.refresh( )
			self.shared.mouse_timer.delegate( self.timed_mover )
			self.shared.mouse_timer.start( 10, 0 )

	def mouse_release_event( self, event ):
		event.accept( )
		self.jog_x = 0
		self.diff_x = 0

	def wheel_event( self, event ):
		return

	def left_button( self ):
		if self.shared.state.get_active_media( ) is None: return
		self.offset += 20
		self.shared.reel.player.refresh( )

	def right_button( self ):
		if self.shared.state.get_active_media( ) is None: return
		self.offset -= 20
		self.shared.reel.player.refresh( )

	# Check that the offset is valid
	def calculate_bounds( self ):
		reel = self.shared.state.get_active_reel( )
		if self.shared.reel.current_h != self.shared.player.height:
			index = reel.find( self.shared.state.get_active_media( ) )
			self.shared.reel.current_w = self.shared.reel.max_w * self.shared.player.height / self.shared.reel.max_h
			self.offset = - index * ( self.shared.reel.current_w + self.shared.reel.gutter )
			self.shared.reel.current_h = self.shared.player.height
		full_width = len( reel.playlist ) * ( self.shared.reel.current_w + self.shared.reel.gutter )
		if self.shared.player.width > full_width:
			self.offset = 0
		elif self.offset < - full_width + self.shared.player.width:
			self.offset = - full_width + self.shared.player.width
		elif self.offset > 0:
			self.offset = 0

#
# Asset mode handler
#

class asset_handler:
	# Constructor
	def __init__( self, shared ):
		self.shared = shared
		self.board = jahwidgets.create_image( self.shared.scribbler.create( 1000, 200 ) )
		self.board.zoom_active = True
		self.dirty = False

	def activate( self ):
		self.shared.state.player.loop = True
		self.shared.player.clear( )

		active = self.shared.state.get_active_media( )

		if active is None:
			active = self.shared.state.player.active
			if active is not None:
				reel = active.reel
				asset = active.asset
				if reel is not None and asset is not None:
					active = reel.get_media( guid = asset.guid( ) )
				else:
					active = None

		if active is not None:
			self.shared.state.prepare_media( active )
			self.shared.player.push_back( active )
			self.shared.player.push_back( self.board )
			self.dirty = True

		self.board.active = active

		self.shared.player.refresh( )

	def accept_drop( self, event ):
		return True

	def process_drop( self, event ):
		self.shared.state.process_media( event )

	def notify_did_resize( self, w, h ):
		pass

	def media_needed( self, media ):
		return True

	# Place the media item in the reel
	def media_geometry( self, media, width, height ):
		self.calculate_bounds( )
		if media is self.board:
			if media.active.asset is not None and self.dirty == True:
				asset = media.active.asset
				media.image = self.shared.scribbler.fill( media.image, 0, 0, 0, 255 )
				text = self.asset_report( asset )
				self.shared.scribbler.set_fg( 255, 255, 255, 255 )
				self.shared.scribbler.set_bg( 0, 0, 0, 255 )
				media.image = self.shared.scribbler.render_text( media.image, str(text), "Arial", 10, jahwidgets.scribbler.alignment.align_left, 0, 0 )
				self.dirty = False

			media.set_geometry( self.shared.reel.current_w, 0, width, height )

		elif media is not None and media.reel is not None:
			width, height = int( media.image.height( ) * media.ar( ) ), media.image.height( )
			width, height = utilities.bind_size_to( width, height, self.shared.reel.max_w, self.shared.reel.max_h )

			reel = media.reel
			if media.dirty and media.image is not None and reel.get_thumb( media ) is None:
				media.image = self.shared.scribbler.rescale( media.image, width, height )
				reel.set_thumb( media.image, media )

			width, height = utilities.bind_size_to( width, height, self.shared.reel.current_w, self.shared.reel.current_h )
			x = ( self.shared.reel.current_w - width ) / 2
			y = ( self.shared.reel.current_h - height ) / 2

			media.t = 1.0
			media.set_geometry( x, y, width, height )

	# Allows caching
	def media_rendered( self, media ):
		# Free up resources for all but the active media item
		if not media.dirty:
			media.close( )

	def mouse_press_event( self, event ):
		event.accept( )

	def mouse_move_event( self, event ):
		event.accept( )

	def mouse_release_event( self, event ):
		event.accept( )

	def wheel_event( self, event ):
		return

	def left_button( self ):
		media = self.shared.state.get_active_media( )
		if media is not None and media.reel is not None and media.asset is not None:
			asset = media.asset
			reel = media.reel
			index = reel.find( guid = asset.guid( ) )
			if index != len( reel.playlist ):
				index = index - 1
				if index < 0: index = len( reel.playlist ) - 1
				self.shared.state.set_active_media( reel.get_media( index ) )

	def right_button( self ):
		media = self.shared.state.get_active_media( )
		if media is not None and media.reel is not None and media.asset is not None:
			asset = media.asset
			reel = media.reel
			index = reel.find( guid = asset.guid( ) )
			if index != len( reel.playlist ):
				index = index + 1
				if index >= len( reel.playlist ): index = 0
				self.shared.state.set_active_media( reel.get_media( index ) )

	def calculate_bounds( self ):
		if self.shared.reel.current_h != self.shared.player.height:
			self.shared.reel.current_w = self.shared.reel.max_w * self.shared.player.height / self.shared.reel.max_h
			self.shared.reel.current_h = self.shared.player.height

	def asset_report( self, asset ):
		text = ""
		for metadata in asset:
			if not metadata.first.startswith( "jahplayer::" ):
				text += ( metadata.first ) + ": " + ( metadata.second ) + "\n"
		return text

#
# Timeline handler
#

class timeline_handler:
	class geometry:
		def __init__( self, offset, width ):
			self.offset = offset
			self.width = width

	class duration:
		def __init__( self, frames, fps ):
			self.frames = frames
			self.fps = fps

	# Constructor
	def __init__( self, shared ):
		self.debug = False

		self.shared = shared
		self.jog_x = 0
		self.diff_x = 0
		self.moving = None
		self.timed_mover = timed_mover( self )
		self.offset = 0
		self.total_width = 0
		self.width = 0
		self.min_image_width = 10 # pixels
		self.proportional = False # use proportional widths

		# Dictionary of image -> geometries
		self.images = {}

		# Cache reel media to avoid unnecessary rebuilds
		self.cached_reel_media = []

	def activate( self ):
		self.shared.state.player.loop = False
		self.shared.player.clear( )
		reel = self.shared.state.get_active_reel( )

		# Get current media list
		media = []
		for i in reel.set:
			m = reel.get_media( guid = i.guid( ) )
			if m is not None:
				media.append( m )

		# Build the timeline images representing our media
		self.rebuild_timeline( media )

		# Now we have our image append to player
		for image in self.images:
			self.shared.state.prepare_media( image )
			self.shared.player.push_back( image )

		self.shared.player.refresh( )

	def accept_drop( self, event ):
		return True

	def process_drop( self, event ):
		self.shared.state.process_media( event )

	def notify_did_resize( self, w, h ):
		# Force a rebuild of all our images
		self.cached_reel_media = []
		self.activate()

	def media_needed( self, media ):
		if media in self.images is False:
			print "Warning: media not in self.images"
			return False

		self.calculate_bounds()
		start = self.images[media].offset + self.offset
		end = start + self.images[media].width

		return (start >= 0 and start < self.shared.player.width) or (end >= 0 and end < self.shared.player.width)

	# Place the media item in the reel
	def media_geometry( self, media, width, height ):
		if media in self.images is False:
			print "Warning: media not in self.images"
			return

		# Fetch image geometry from our cache
		width = int(self.images[media].width)
		height = int(self.shared.reel.current_h)

		# Show everything except the active media as semi-transparent
		media.t = 0.5
		active_media = self.shared.state.get_active_media( )
		if active_media and active_media.asset and active_media.asset.guid() == media.asset.guid(): 
			media.t = 1.0

		# Set geometry
		x = self.offset + self.images[media].offset
		y = 0

		media.set_geometry( int(x), int(y), int(width), int(height) )

	# Allows caching
	def media_rendered( self, media ):
		# Free up resources for all but the active media item
		if not media.dirty:
			media.close( )

	def mouse_press_event( self, event ):
		event.accept( )
		if event.isDoubleClick:
			media = self.shared.player.find( event.x, event.y )
			# Find the corresponding 'reel' item
			if media and media.asset:
				reel = self.shared.state.get_active_reel( )
				reel_media = reel.get_media( guid = media.asset.guid() )
				if reel_media is not None:
					self.shared.state.set_active_media( reel_media )
		self.jog_x = event.x
		self.diff_x = 0

	def mouse_move_event( self, event ):
		event.accept( )
		if self.jog_x:
			if event.x - self.jog_x:
				self.diff_x = event.x - self.jog_x
			self.offset = self.offset + self.diff_x
			if event.x > 0: self.jog_x = event.x
			self.shared.player.refresh( )
			self.shared.mouse_timer.delegate( self.timed_mover )
			self.shared.mouse_timer.start( 10, 0 )

	def mouse_release_event( self, event ):
		event.accept( )
		self.jog_x = 0
		self.diff_x = 0

	def wheel_event( self, event ):
		pass

	def left_button( self ):
		#if self.shared.state.get_active_media( ) is None: return
		self.offset += 20
		self.shared.reel.player.refresh( )

	def right_button( self ):
		#if self.shared.state.get_active_media( ) is None: return
		self.offset -= 20
		self.shared.reel.player.refresh( )

	# Check that the offset is valid
	def calculate_bounds( self ):
		reel = self.shared.state.get_active_reel( )
		if self.shared.reel.current_h != self.shared.player.height:
			self.shared.reel.current_w = self.shared.reel.max_w * self.shared.player.height / self.shared.reel.max_h
			self.shared.reel.current_h = self.shared.player.height

		if self.shared.player.width > self.total_width:
			self.offset = 0
		elif self.offset < - self.total_width + self.shared.player.width:
			self.offset = - self.total_width + self.shared.player.width
		elif self.offset > 0:
			self.offset = 0

	# Rebuild the offsets and widths of all our media on the current reel's timeline
	def rebuild_timeline( self, media ):
		# Check there is something to do
		if media == self.cached_reel_media and self.shared.player.width == self.width:
			return

		self.cached_reel_media = [] # Rebuilt below
		self.width = self.shared.player.width

		self.calculate_bounds()
		
		# Get the durations
		durations = []
		for m in media:
			durations.append( self.media_duration( m ) )
			self.cached_reel_media.append( m )

		if self.proportional:
			pixelDurations, self.total_width = self.build_proportional_geometries( durations, self.shared.player.width )
		else:
			pixelDurations, self.total_width = self.build_nonproportional_geometries( durations, self.shared.player.width )

		# Now accumulate the offsets and build the images
		i = 0
		offset = 0
		self.images = {}

		for width in pixelDurations:
			geometry = self.geometry( offset, width )

			# Build the new image
			image = self.shared.scribbler.create( int(geometry.width), int(self.shared.reel.current_h) )
			image = self.shared.scribbler.fill( image, 30, 144, 255, 255 ) # light blue
			image = self.shared.scribbler.border( image, 1, 0, 0, 0, 255 )

			# And add text if there's room
			filename = str( media[i].asset.find( "name" ).second )
			length = "\n" + self.shared.scribbler.timecode( durations[i].frames, durations[i].fps, False )
			label = filename + length

			bounds = self.shared.scribbler.bounds_for_text( label, "Arial", 11, jahwidgets.scribbler.alignment.align_left )
			lengthBounds = self.shared.scribbler.bounds_for_text( length, "Arial", 11, jahwidgets.scribbler.alignment.align_left )

			usableWidth = geometry.width - 4

			# Remove the length string entirely if it won't fit
			if lengthBounds.width > usableWidth: length = ""

			# Eat away the filename until it fits...
			while bounds.width > usableWidth:
				filename = filename[:-1]
				if len(filename) is 0: break
				label = filename + "..." + length
				bounds = self.shared.scribbler.bounds_for_text( label, "Arial", 11, jahwidgets.scribbler.alignment.align_left )

			if bounds.width <= usableWidth and bounds.height <= self.shared.reel.current_h:
				self.shared.scribbler.save()
				self.shared.scribbler.set_fg( 255, 255, 255, 255 )
				# Anti-alias to black for now, other colors are doing stranges things when composited...
				self.shared.scribbler.set_bg( 0, 0, 0, 255 )
				dx = int( geometry.width - bounds.width ) / 2
				dy = int( self.shared.reel.current_h - bounds.height ) / 2
				image = self.shared.scribbler.render_text( image, label, "Arial", 11, jahwidgets.scribbler.alignment.align_left, dx, dy )
				self.shared.scribbler.restore()

			m = jahwidgets.create_image( image )
			m.asset = media[i].asset # Store the asset on our thumbnail so we can highlight the active asset etc
			self.images[ m ] = geometry

			i += 1
			offset += width

		# DEBUG
		if self.debug:
			print "Printing pixel widths:"
			i = 0
			for geometry in self.images.values():
				print str(durations[i].frames) + " -> " + str(geometry.width)
				i += 1

			print "Total width %d" % self.shared.player.width
			print "Clip width %d" % self.total_width
		# DEBUG

	def build_proportional_geometries( self, durations, maxWidth ):
		if len(durations) == 0: return [], 0

		# Find min duration
		minDuration = 1.e99
		maxDuration = 0.
		for duration in durations:
			if duration.frames < minDuration: minDuration = duration.frames
			if duration.frames > maxDuration: maxDuration = duration.frames

		if minDuration is 0.:
			print "Warning: zero length clip found"
			minDuration = 1

		pixelsPerDuration = float(self.min_image_width) / minDuration

		pixelDurations = []
		totalPixelWidth = 0

		# Metric used to increase width used when many clips of similar length are short
		shortWidth = 50
		shortWidthCount = 0

		for duration in durations:
			pixelDurations.append( float(duration.frames) * pixelsPerDuration )
			if pixelDurations[-1] <= shortWidth: shortWidthCount += 1
			totalPixelWidth += pixelDurations[-1]

		def scaleDurations( factor ):
			shortCount = 0
			totalWidth = 0
			for i in range(len(pixelDurations)):
				pixelDurations[i] = factor * pixelDurations[i]
				if pixelDurations[i] <= shortWidth: shortCount += 1
				totalWidth += pixelDurations[i]
			return totalWidth, shortCount

		# scale up to at least fill the width
		if totalPixelWidth < maxWidth:
			totalPixelWidth, shortWidthCount = scaleDurations( float(maxWidth) / totalPixelWidth )

		# If we still have too many short widths then double the available space
		if (float(maxDuration) / minDuration) <= 2.0:
			# Scale up by factor of two
			while (float(shortWidthCount) / len(durations)) > 0.3:
				totalPixelWidth, shortWidthCount = scaleDurations( 2.0 )

		return pixelDurations, totalPixelWidth


	def build_nonproportional_geometries( self, durations, maxWidth ):
		# TODO: Doesn't perform too well if there is a large ratio for longest/shortest clip
		if len(durations) == 0: return [], 0

		# how high the ratio of longestDuration/shortedDuration minWidth clips can get before we double the available width
		maxClippedDurationRatio = 5. 

		# Find total duration
		totalDuration = 0.
		for duration in durations:
			totalDuration = totalDuration + duration.frames

		durationPer10Pixels = 10. * totalDuration / maxWidth

		# Run over all clips and assign a pixel width based on durationPer10Pixels
		pixelDurations = []
		totalPixelWidth = 0.
		minClipCount = 0
		minClippedDuration = 1.e99
		maxClippedDuration = 0.

		for duration in durations:
			if duration.frames <= durationPer10Pixels: 
				pixelDurations.append( self.min_image_width )
				minClipCount += 1
				if duration.frames < minClippedDuration: minClippedDuration = duration.frames
				if duration.frames > maxClippedDuration: maxClippedDuration = duration.frames
			else: 
				pixelDurations.append( 10. * duration.frames / durationPer10Pixels )

			totalPixelWidth += pixelDurations[-1]

		if totalPixelWidth == maxWidth:
			return pixelDurations, totalPixelWidth

		if totalPixelWidth < maxWidth:
			factor = 1. * maxWidth / totalPixelWidth
			# Scale all items up
			for i in range(len(pixelDurations)):
				pixelDurations[i] = factor * pixelDurations[i]

			# And we're done
			return pixelDurations, maxWidth

		while totalPixelWidth > maxWidth:
			# If we have too many short clips just double the maxWidth
			# and start again. This condition may need some tweaking to get right
			totalMinClipWidth = minClipCount * self.min_image_width
			# if totalMinClipWidth > (0.5 * maxWidth):
			if maxClippedDuration / minClippedDuration > maxClippedDurationRatio: 
				return self.build_nonproportional_geometries( durations, maxWidth + self.shared.player.width )

			if totalPixelWidth == totalMinClipWidth: break

			# Else scale clips > 10px down
			factor = (maxWidth - totalMinClipWidth) / (totalPixelWidth - totalMinClipWidth)
			totalPixelWidth = 0.
			minClipCount = 0

			for i in range(len(pixelDurations)):
				# scale
				if pixelDurations[i] > self.min_image_width: pixelDurations[i] = factor * pixelDurations[i]

				# no smaller than minWidth
				if pixelDurations[i] <= self.min_image_width: 
					pixelDurations[i] = self.min_image_width
					minClipCount += 1

				totalPixelWidth += pixelDurations[i]

		# Done
		return pixelDurations, totalPixelWidth

	def media_duration( self, media ):
		if media.asset is not None and str( media.asset.find( "frames" ).first ) != "":
			frames = int( str( media.asset.find( "frames" ).second ) )
			if str( media.asset.find( "fps" ).first ) != "":
				return self.duration( frames, float(media.asset.find( "fps" ).second) )
			else:
				return self.duration( frames, 25 )
		else:
			return self.duration(1, 25)

		
#
# Timer delegate for rendering stuff
#

class timed_render( jahwidgets.timer_delegate ):
	def __init__( self, handler ):
		jahwidgets.timer_delegate.__init__( self )
		self.handler = handler
		self.active = None
		self.index = 0
		self.gap = 0
		self.count = 0

	def timer( self ):
		if self.index * self.gap >= self.active.length: 
			if self.handler.shared.state.get_active_mode( ) == "mode_filmstrip":
				self.handler.shared.state.player.speed = self.handler.orig_speed
			return False
		self.handler.shared.state.player.speed = 0
		self.active.seek( self.index * self.gap, False )
		image = self.active.fetch_image( )
		if image is not None:
			width, height = int( image.height( ) * self.active.ar( ) ), image.height( )
			width, height = utilities.bind_size_to( width, height, self.handler.shared.reel.max_w, self.handler.shared.reel.max_h )
			image = self.handler.shared.scribbler.rescale( image, width, height )
			media = jahwidgets.create_image( image )
			media.asset = self.active.asset
			media.pos = self.index * self.gap
			self.handler.shared.state.prepare_media( media )
			self.handler.cached_thumbs[ media ] = self.index * self.gap
			self.index += 1
			if self.handler.shared.state.get_active_mode( ) == "mode_filmstrip":
				self.handler.activate( )
				if self.index > self.count: 
					self.handler.shared.state.player.speed = self.handler.orig_speed
			return self.index <= self.count
		else:
			return False

#
# Filmstrip mode handler
#

class filmstrip_handler:
	# Constructor
	def __init__( self, shared ):
		self.shared = shared
		self.active_media = None
		self.cloned_media = None
		self.cached_thumbs = { }
		self.timed_render = timed_render( self )
		self.timed_mover = timed_mover( self )
		self.offset = 0
		self.jog_x = 0
		self.diff_x = 0

	def activate( self ):
		self.shared.state.player.loop = True
		self.shared.player.active = None
		self.shared.player.clear( )

		active = self.shared.state.player.active
		if active is None or active.reel is None: 
			self.shared.player.refresh( )
			return

		reel = active.reel

		if self.active_media is not active:
			self.orig_speed = self.shared.state.player.speed
			self.shared.state.player.speed = 0
			self.active_media = active
			self.cloned_media = jahwidgets.create_media( active.uri( ) )
			self.cloned_media.asset = active.asset
			self.cloned_media.open( )
			self.cached_thumbs = { }
			self.gap = active.length / 20
			if self.gap == 0: self.gap = 1

			self.timed_render.gap = self.gap
			self.timed_render.active = self.cloned_media
			self.timed_render.index = 0
			self.timed_render.count = 20

			self.shared.render_timer.delegate( self.timed_render )
			self.shared.render_timer.start( 0, 0 )

		for media in self.cached_thumbs:
			self.shared.state.prepare_media( media )
			self.shared.player.push_back( media )

		self.shared.player.refresh( )

	def accept_drop( self, event ):
		return True

	def process_drop( self, event ):
		self.shared.state.process_media( event )

	def notify_did_resize( self, w, h ):
		pass

	def media_needed( self, media ):
		self.calculate_bounds( )
		return True

	# Place the media item in the reel
	def media_geometry( self, media, width, height ):
		index = self.cached_thumbs[ media ] / self.gap
		width, height = utilities.bind_size_to( width, height, self.shared.reel.current_w, self.shared.reel.current_h )
		x = index * ( self.shared.reel.current_w + self.shared.reel.gutter ) + ( self.shared.reel.current_w - width ) / 2
		y = ( self.shared.reel.current_h - height ) / 2
		media.set_geometry( self.offset + x, y, width, height )

	# Allows caching
	def media_rendered( self, media ):
		pass

	def mouse_press_event( self, event ):
		event.accept( )
		self.jog_x = event.x
		self.diff_x = 0

		if event.isDoubleClick:
			media = self.shared.player.find( event.x, event.y )
			if media is not None:
				self.shared.state.position_prop.set( media.pos )
				self.shared.state.player.refresh( )

	def mouse_move_event( self, event ):
		event.accept( )
		if self.jog_x:
			if event.x - self.jog_x:
				self.diff_x = event.x - self.jog_x
			self.offset = self.offset + self.diff_x
			if event.x > 0: self.jog_x = event.x
			self.shared.player.refresh( )
			self.shared.mouse_timer.delegate( self.timed_mover )
			self.shared.mouse_timer.start( 10, 0 )

	def mouse_release_event( self, event ):
		event.accept( )
		self.jog_x = 0
		self.diff_x = 0

	def wheel_event( self, event ):
		pass

	def left_button( self ):
		pass

	def right_button( self ):
		pass

	def calculate_bounds( self ):
		active = self.shared.state.player.active
		reel = active.reel
		if self.shared.reel.current_h != self.shared.player.height:
			self.shared.reel.current_w = self.shared.reel.max_w * self.shared.player.height / self.shared.reel.max_h
			self.shared.reel.current_h = self.shared.player.height
		full_width = len( self.cached_thumbs ) * ( self.shared.reel.current_w + self.shared.reel.gutter )
		if self.shared.player.width > full_width:
			self.offset = 0
		elif self.offset < - full_width + self.shared.player.width:
			self.offset = - full_width + self.shared.player.width
		elif self.offset > 0:
			self.offset = 0

#
# Drag n Drop delegate
#

class drop_delegate( jahwidgets.drop_delegate ):
	def __init__( self, shared, handlers ):
		jahwidgets.drop_delegate.__init__( self )
		self.shared = shared
		self.handlers = handlers

	def accept( self ):
		return self.handlers[ self.shared.state.get_active_mode( ) ].accept_drop( self )

	def operate( self ):
		return self.handlers[ self.shared.state.get_active_mode( ) ].process_drop( self )

#
# Geometry delegate
#

class geometry_delegate( jahwidgets.player_delegate ):
	# Constructor
	def __init__( self, shared, handlers ):
		jahwidgets.player_delegate.__init__( self )
		self.shared = shared
		self.handlers = handlers

	def media_needed( self, media ):
		return self.handlers[ self.shared.state.get_active_mode( ) ].media_needed( media )

	# Place the media item in the reel
	def media_geometry( self, media, width, height ):
		self.handlers[ self.shared.state.get_active_mode( ) ].media_geometry( media, width, height )

	# Allows caching
	def media_rendered( self, media ):
		self.handlers[ self.shared.state.get_active_mode( ) ].media_rendered( media )

	def notify_did_resize( self, w, h ):
		self.handlers[ self.shared.state.get_active_mode( ) ].notify_did_resize( w, h )


#
# Mouse delegate
#

class mouse_delegate( jahwidgets.mouse_delegate ):
	def __init__( self, shared, handlers ):
		jahwidgets.mouse_delegate.__init__( self )
		self.shared = shared
		self.handlers = handlers

	def mousePressEvent( self ):
		self.handlers[ self.shared.state.get_active_mode( ) ].mouse_press_event( self )

	def mouseMoveEvent( self ):
		self.handlers[ self.shared.state.get_active_mode( ) ].mouse_move_event( self )

	def mouseReleaseEvent( self ):
		self.handlers[ self.shared.state.get_active_mode( ) ].mouse_release_event( self )

#
# Wheel delegate
#

class wheel_delegate( jahwidgets.wheel_delegate ):
	def __init__( self, shared, handlers ):
		jahwidgets.wheel_delegate.__init__( self )
		self.shared = shared
		self.handlers = handlers

	def wheelEvent( self ):
		self.handlers[ self.shared.state.get_active_mode( ) ].wheel_event( self )

#
# Left scroll button observer
#

class left_observer( openpluginlib.observer ):
	def __init__( self, shared, handlers ):
		openpluginlib.observer.__init__( self )
		self.shared = shared
		self.handlers = handlers

	def updated( self, subject ):
		self.handlers[ self.shared.state.get_active_mode( ) ].left_button( )

#
# Right scroll button observer
#

class right_observer( openpluginlib.observer ):
	def __init__( self, shared, handlers ):
		openpluginlib.observer.__init__( self )
		self.shared = shared
		self.handlers = handlers

	def updated( self, subject ):
		self.handlers[ self.shared.state.get_active_mode( ) ].right_button( )

#
# Mode observer
#

class mode_observer( openpluginlib.observer ):
	def __init__( self, state, mode ):
		openpluginlib.observer.__init__( self )
		self.state = state
		self.mode = mode

	def updated( self, subject ):
		self.state.set_active_mode( self.mode )

#
# The reel widget
#

class widget:
	def __init__( self, parent, state ):
		# Store the application state
		self.state = state

		# Create the ui
		self.handle = jahwidgets.create_ui_from_file( utilities.get_absolute_path( "reel.ui" ), parent )

		# Set up our player instance
		self.player_handle = jahwidgets.find( "player", self.handle )
		self.player = jahwidgets.player( self.player_handle )
		self.player.show_tips = True

		# The size of each item
		self.max_w = 100
		self.max_h = 60
		self.gutter = 2
		self.current_h = self.player.height
		self.current_w = self.max_w * self.player.height / self.max_h

		# Construct the handler objects for each mode
		self.shared = shared_handler( self )
		self.handlers = { }
		self.handlers[ "mode_reel" ] = reel_handler( self.shared )
		self.handlers[ "mode_asset" ] = asset_handler( self.shared )
		self.handlers[ "mode_filmstrip" ] = filmstrip_handler( self.shared )
		self.handlers[ "mode_timeline" ] = timeline_handler( self.shared )
		self.handlers[ "mode_playlist" ] = reel_handler( self.shared )

		# Create our delegates
		self.dnd = drop_delegate( self.shared, self.handlers )
		self.geometry = geometry_delegate( self.shared, self.handlers )
		self.mouse = mouse_delegate( self.shared, self.handlers )
		self.wheel = wheel_delegate( self.shared, self.handlers )

		# Register the delegates
		self.player.register_drop_delegate( self.dnd )
		self.player.register_player_delegate( self.geometry )
		self.player.register_mouse_delegate( self.mouse )
		self.player.register_wheel_delegate( self.wheel )

		# Construct properties for the left and right buttons
		self.left_prop = openpluginlib.property( openpluginlib.createKeyFromString( "left" ) )
		self.left_prop.setAlwaysNotify( True )
		jahwidgets.connect( "left:clicked", self.handle, self.left_prop )
		self.left_prop.attach( left_observer( self.shared, self.handlers ) )

		self.right_prop = openpluginlib.property( openpluginlib.createKeyFromString( "right" ) )
		self.right_prop.setAlwaysNotify( True )
		jahwidgets.connect( "right:clicked", self.handle, self.right_prop )
		self.right_prop.attach( right_observer( self.shared, self.handlers ) )

		# Register ourselves for reel modifications
		self.state.register_reel_listener( self )

		# Set up the modes of the widget
		self.mode_reel_prop = self.create_mode_observer( "mode_reel", "pb_reel" )
		self.mode_asset_prop = self.create_mode_observer( "mode_asset", "pb_asset" )
		self.mode_filmstrip_prop = self.create_mode_observer( "mode_filmstrip", "pb_filmstrip" )
		self.mode_timeline_prop = self.create_mode_observer( "mode_timeline", "pb_timeline" )
		self.mode_playlist_prop = self.create_mode_observer( "mode_playlist", "pb_playlist" )

		# populate combobox
		self.cb_reel_prop = utilities.create_property_and_attach( self.handle, "cb_reel:currentItem", utilities.call_method_observer( self.reel_selected ) )
		self.cb_reel_items_prop = openpluginlib.property( openpluginlib.createKeyFromString( "reel_items" ) )
		jahwidgets.connect( "cb_reel:items", self.handle, self.cb_reel_items_prop )
		self.cb_reel_name_edited_prop = utilities.create_property_and_attach( self.handle, "cb_reel:textChanged", utilities.call_method_observer( self.reel_name_edited ) )

		# Register ourselves as a mode listener and set the initial state
		self.mode = None
		self.state.register_mode_listener( self )
		self.state.set_active_mode( "mode_reel" )

	# Creates a mode observer
	def create_mode_observer( self, name, button ):
 		property = openpluginlib.property( openpluginlib.createKeyFromString( name ) )
		property.setAlwaysNotify( True )
		jahwidgets.connect( button + ":clicked", self.handle, property )
		property.attach( mode_observer( self.state, name ) )
		return property

	# Callback for mode change
	def notify_mode_change( self, mode ):
		if self.mode is None or mode != self.mode:
			self.mode = mode
			self.notify_reel_change( )

	# Callback for reel change
	def notify_reel_change( self ):
		if not self.cb_reel_items_prop.valueAsWStringList() == self.state.reel_names:
			jahwidgets.set_value( "cb_reel:items", self.handle, openpluginlib.any( self.state.reel_names ) )
		# TODO: Need to migrate the name of the reel into oal?
		reel = self.state.get_active_reel( )
		path, name = os.path.split( reel.filename )
		self.cb_reel_prop.set( self.state.reel_names.index( name ) )
		self.handlers[ self.shared.state.get_active_mode( ) ].activate( )

	def reel_selected( self, subject ):
		name = self.state.reel_names[ self.cb_reel_prop.valueAsInt() ]
		i = self.state.reel_names.index( name )
		if self.shared.state.get_active_reel( ) is not self.shared.state.reels[ i ]:
			self.shared.state.set_active_reel( self.shared.state.reels[ i ] )

	def reel_name_edited( self, subject ):
		print "reel_name_edited"
